#ifndef ATLFast_H
#define ATLFast_H

//////////////////////////////////////////////////////////////////////////
//                                                                      //
// ATLFast                                                              //
//                                                                      //
// Main class to control the ATLFast program.                           //
//                                                                      //
// This class :                                                         //
//   - Initialises the run default parameters                           //
//   - Provides API to Set/Get run parameters                           //
//   - Creates the support lists (TClonesArrays) for the Event structure//
//   - Creates the physics objects makers                               //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#ifndef ROOT_TTree
#include <TTree.h>
#endif
#ifndef ROOT_TList
#include <TList.h>
#endif
#ifndef ROOT_TMath
#include <TMath.h>
#endif
#ifndef ATLFHistBrowser_H
#include "ATLFHistBrowser.h"
#endif
#ifndef ATLFBigBang_H
#include "ATLFBigBang.h"
#endif
#ifndef ATLFMaker_H
#include "ATLFMaker.h"
#endif

class TBrowser;
class TChain;
class ATLFMCMaker;
class ATLFClusterMaker;
class ATLFElectronMaker;
class ATLFMuonMaker;
class ATLFPhotonMaker;
class ATLFJetMaker;
class ATLFTrackMaker;
class ATLFTriggerMaker;
class ATLFMiscMaker;
class ATLFVirtualDisplay;

class ATLFast : public TNamed {

private:
   Int_t               m_Version;           //ATLFast version number
   Int_t               m_VersionDate;       //ATLFast version date
   Int_t               m_Run;               //Run number
   Int_t               m_Event;             //Event number
   Int_t               m_Mode;              //Run mode
   TTree              *m_Tree;              //!Pointer to the Root tree
   TList              *m_Makers;            //List of Makers
//    pointers to standard Makers
   ATLFMCMaker        *m_MCMaker;           //Pointer to MCMaker
   ATLFClusterMaker   *m_ClusterMaker;      //Pointer to ClusterMaker
   ATLFElectronMaker  *m_ElectronMaker;     //Pointer to ElectronMaker
   ATLFMuonMaker      *m_MuonMaker;         //Pointer to MuonMaker
   ATLFPhotonMaker    *m_PhotonMaker;       //Pointer to PhotonMaker
   ATLFJetMaker       *m_JetMaker;          //Pointer to JetMaker
   ATLFTrackMaker     *m_TrackMaker;        //Pointer to TrackMaker
   ATLFTriggerMaker   *m_TriggerMaker;      //Pointer to TriggerMaker
   ATLFMiscMaker      *m_MiscMaker;         //Pointer to MiscMaker
//    flags and switches
   Int_t               m_Luminosity;        //Luminosity option (low=1, high=2)
   Bool_t              m_Bfield;            //B-field (on=1,off=0)
   Bool_t              m_Smearing;          //Smearing on=1, off=0
   Int_t               m_SmearMuonOpt;      //Smear for muons: muon=1, track=2,combin=3
   Int_t               m_SmearMuonFun;      //Smear for muons: exact=1, approxim=2
   Int_t               m_SUSYcodeLSP;       //Code for SUSY LSP particle
   Int_t               m_TrackFinding;      //Track/finding on=1,off=0
   ATLFHistBrowser     m_HistBrowser;       //Object to Browse Maker histograms
   ATLFBigBang         m_BigBang;           //!Object to Browse generated particles
   ATLFVirtualDisplay *m_Display;           //!Pointer to Event display object

public:
                      ATLFast();
                      ATLFast(const char *name, const char *title="The ATLAS fast MonteCarlo");
   virtual           ~ATLFast();
   virtual void       Browse(TBrowser *b);
   virtual void       Draw(Option_t *option="");  // *MENU*
   Int_t              GetVersion() {return m_Version;}
   Int_t              GetVersionDate() {return m_VersionDate;}
   virtual void       Clear(Option_t *option="");
   ATLFVirtualDisplay *Display() {return m_Display;}
   virtual void       FillClone();
   virtual void       Finish();
   virtual void       GetEvent(Int_t event=1);  // *MENU*
   virtual void       Init();
   Bool_t             IsFolder() const {return kTRUE;}
   virtual Int_t      Make(Int_t i=0);
   virtual void       Paint(Option_t *option="");
   virtual void       PrintInfo();
   virtual void       SetDefaultParameters();

   TList             *Makers()    {return m_Makers;}
   ATLFMaker         *Maker(const char *name) {return (ATLFMaker*)m_Makers->FindObject(name);}
   ATLFMCMaker       *MCMaker()       {return m_MCMaker;}
   ATLFClusterMaker  *ClusterMaker()  {return m_ClusterMaker;}
   ATLFElectronMaker *ElectronMaker() {return m_ElectronMaker;}
   ATLFMuonMaker     *MuonMaker()     {return m_MuonMaker;}
   ATLFPhotonMaker   *PhotonMaker()   {return m_PhotonMaker;}
   ATLFJetMaker      *JetMaker()      {return m_JetMaker;}
   ATLFTrackMaker    *TrackMaker()    {return m_TrackMaker;}
   ATLFTriggerMaker  *TriggerMaker()  {return m_TriggerMaker;}
   ATLFMiscMaker     *MiscMaker()     {return m_MiscMaker;}
   TTree             *Tree() {return m_Tree;}

   Int_t             Run()   {return m_Run;}
   Int_t             Event() {return m_Event;}
   Int_t             Mode()  {return m_Mode;}

//    Getters for flags and switches
   Int_t             Luminosity()   {return m_Luminosity;}
   Bool_t            Bfield()       {return m_Bfield;}
   Bool_t            Smearing()     {return m_Smearing;} 
   Int_t             SmearMuonOpt() {return m_SmearMuonOpt;}
   Int_t             SmearMuonFun() {return m_SmearMuonFun;}
   Int_t             SUSYcodeLSP()  {return m_SUSYcodeLSP;}
   Bool_t            TrackFinding() {return m_TrackFinding;}


//    Setter for Event Display
   void           SetDisplay(ATLFVirtualDisplay *display) {m_Display=display;}
//    Setters for flags and switches
   void           SetLuminosity(Int_t lumi=1)   {m_Luminosity=lumi;}
   void           SetBfield(Bool_t field=1)     {m_Bfield=field;}
   void           SetSmearing(Bool_t val=1)     {m_Smearing=val;} 
   void           SetSmearMuonOpt(Int_t val=3)  {m_SmearMuonOpt=val;}
   void           SetSmearMuonFun(Int_t val=2)  {m_SmearMuonFun=val;}
   void           SetSUSYcodeLSP(Int_t val=66)  {m_SUSYcodeLSP=val;}
   void           SetTrackFinding(Bool_t val=0) {m_TrackFinding=val;}

   virtual void   SetRun(Int_t run=1)     {m_Run=run;}
   virtual void   SetEvent(Int_t event=1) {m_Event=event;}
   virtual void   SetMode(Int_t mode=0)   {m_Mode=mode;}

   void           FillTree();
   void           InitChain(TChain *chain);
   void           MakeTree(const char* name="T", const char*title="ATLFast tree");

   void           SortDown(Int_t n, Float_t *a, Int_t *index, Bool_t down=kTRUE);

   ClassDef(ATLFast, 1)   //ATLFast control class
};

R__EXTERN ATLFast *gATLFast;

#endif
